/**
* Name: KissNightclubwithNorms
* Author: mathieu
* Description: 
* Tags: Tag1, Tag2, TagN
*/

model KissNightclubwithNorms

global{
	//Shapefile of the walls
	file wall_shapefile <- shape_file("../includes/murs.shp");
	//Shapefile of the exit
	file exit_shapefile <- shape_file("../includes/sortie.shp");
	
	file fire_shapefile <- shape_file("../includes/feu.shp");
	file wc_shapefile <- shape_file("../includes/wc.shp");
	
	geometry shape <- envelope(wall_shapefile);

	//DImension of the grid agent
	int nb_cols <- 100;
	int nb_rows <- 100;
	
	int nb_victims <- 1200 parameter: true;
	float vision_amplitude <- 60.0 max: 180.0 parameter: true;
	bool use_masked_by <- true parameter: true;
	bool use_congestion <- false parameter: true;
	float coeff_congestion <- 100.0 parameter: true;
	float dist_congestion <-1.0 parameter: true;
	float mean_know_exitDoor <- 90.0 parameter: true;
	float std_know_exitDoor <- 30.0 parameter: true;
	
	float min_smoke <- 30.0 parameter: true;
	float max_smoke <- 60.0 parameter: true;
	
	float mean_speed <- 1.5 parameter: true;
	float std_speed <- 0.5 parameter: true;
	
	float mean_persistence <- 0.8 parameter: true;
	float std_persistence <- 0.2 parameter: true;
	
	float riskAversion <- 0.75 parameter: true;
	float uncertaintyConversion <- 0.25 parameter: true;
	float contagionThreshold <- 0.5 parameter: true;
	
	float propFriend <- 0.5;
	float thresholdFriend <- 0.25;
	
	//perception distance
	float normal_distance <- 4.0 parameter: true;
	float distance_in_smoke <- 1.5 parameter: true;
	int nbDead <- 0;
	int nbInjured <- 0;
	
	list<cell> free_cells;
	
	
	//vocabulary of the model
	predicate exitDoor <- new_predicate("exitLocation");
	predicate directionExitDoor <- new_predicate("direction of the exit");
	
	predicate fireSaw <- new_predicate("fire");
	predicate noFire <- new_predicate("fire",false);
	predicate smokeSaw <- new_predicate("smoke");
	predicate reliableSign <- new_predicate("reliable exit sign");
	predicate fleeing <- new_predicate("flee");
	predicate wandering <- new_predicate("wander");
	predicate helpingFriend <- new_predicate("helpFriend");
	predicate beingHelped <- new_predicate("beingHelped");
	predicate wrongExit <- new_predicate("wrongExit");
	emotion fearFire <- new_emotion("fear",fireSaw);
	
	
	
	init{
	//Creation of the wall and initialization of the cell is_wall attribute
		create wall from: wall_shapefile {
			ask cell overlapping self {
				is_wall <- true;
			}
		}
		free_cells <- cell where not each.is_wall;
		ask cell {
			free_neighbors <- neighbors where not each.is_wall;
		}
		//Creation of the exit and initialization of the cell is_exit attribute
		create exit from: exit_shapefile {
			ask (cell overlapping self) where not each.is_wall{
				is_exit <- true;				
			}			
		}
		
		create bathroom from: wc_shapefile;
		
		
		create fire number:1;
		
		
		create victim number: nb_victims;
		
	}
	
	reflex stop when: length(victim)=0{
		write "nbDead " + nbDead;
		write "nbInjured " + (nbInjured-nbDead);
		write "time : " + average_duration;
		do pause;
//		do halt;
	}
}

//Grid species to discretize space
grid cell width: nb_cols height: nb_rows neighbors: 8 {
	bool is_wall <- false;
	bool is_exit <- false;
	bool is_fire <- false;
	float smoke_level <-0.0 max: 100.0;
	list<cell> free_neighbors;
	rgb color <- #white;	
	
	reflex updateColor{
//		if(smoke_level>10.0){
//			color <-rgb("black",10);
//		}
//		if(smoke_level>20.0){
//			color <-rgb("black",30);
//		}
//		if(smoke_level>30.0){
//			color <-rgb("black",50);
//		}
//		if(smoke_level>40.0){
//			color <-rgb("black",70);
//		}
//		if(smoke_level>50.0){
//			color <-rgb("black",100);
//		}
//		if(smoke_level>60.0){
//			color <-rgb("black",140);
//		}
//		if(smoke_level>70.0){
//			color <-rgb("black",170);
//		}
//		if(smoke_level>80.0){
//			color <-rgb("black",200);
//		}
//		if(smoke_level>90.0){
//			color <-rgb("black",220);
//		}
//		if(smoke_level>=100.0){
//			color <- #black;
//		}
		if(smoke_level>10.0){
			color <-rgb(220,220,220);
		}
		if(smoke_level>20.0){
			color <-rgb(200,200,200);
		}
		if(smoke_level>30.0){
			color <-rgb(170,170,170);
		}
		if(smoke_level>40.0){
			color <-rgb(140,140,140);
		}
		if(smoke_level>50.0){
			color <-rgb(100,100,100);
		}
		if(smoke_level>60.0){
			color <-rgb(70,70,70);
		}
		if(smoke_level>70.0){
			color <-rgb(50,50,50);
		}
		if(smoke_level>80.0){
			color <-rgb(30,30,30);
		}
		if(smoke_level>90.0){
			color <-rgb(10,10,10);
		}
		if(smoke_level>=100.0){
			color <- #black;
		}
	}
	
	reflex ignition when: is_fire{
		smoke_level <- 100.0;
		ask neighbors{
			smoke_level <- smoke_level;
		}
	}
		
	reflex propagationSmoke when: (smoke_level > 0.0){
		if(smoke_level>5.0){
			if(!is_wall){
				ask neighbors{
					if(smoke_level<100 and smoke_level<myself.smoke_level){
						smoke_level<-smoke_level+1.0;
						myself.smoke_level <- myself.smoke_level - 0.5;
					}
				}
			}
			else{
				if(flip(0.12)){
					ask neighbors{
						smoke_level<-smoke_level+1.0;
					}
				}
			}
		}
	}
}
//Species exit which represent the exit
species exit {
	aspect default {
		draw shape color: #blue;
	}
}
//Species which represent the wall
species wall {
	aspect default {
		draw shape color: #black ;
	}
}

species bathroom {
	init{
		location <- (free_cells closest_to self).location;
	}
}

species fire{	
	rgb f_color <- #orange;	
	init {
		location <- cell[10,5].location;
		ask cell[10,5]{
			is_fire <- true;
		}
	}
	
	aspect default {
		draw sphere(1) color: f_color;
	}	
}

species victim skills: [moving] control: /*parallel*/simple_bdi{
	point target;
	rgb color <- #white;
	geometry perceived_area;
	cell myCell update: cell(location);
	path myPath;
	float life <- 100.0;
	bool injured <- false;
	float perceive_distance <- 4.0;
	float smokeQuantity max: 100.0;
	bathroom myBathroom <- bathroom closest_to self;
	
	bool use_emotions_architecture <- true;
	bool use_social_architecture <- true;
	bool use_personality <- true;
	
	float openness <- /*1.0;*/gauss(0.5,0.12);
	float conscientiousness <- /*1.0;*/gauss(0.5,0.12); 
	float extroversion <- /*1.0;*/gauss(0.5,0.12);
	float agreeableness <- /*1.0;*/gauss(0.5,0.12);
	float neurotism <- /*1.0;*/gauss(0.5,0.12);
	
	victim friend <-nil;
	victim perceivedOther <- nil;
	bool hasFriend <- false;
	bool needHelp <- false;
	
	float solidarity<-0.0;
	float thresholdLaw <- 1.0; 
	float thresholdNorm <- gauss(0.5,0.12);
		
	action walk{
		float actual_speed <- speed;
		if (use_congestion) {
			int victims_around <- length(victim at_distance dist_congestion);
			actual_speed <- max([0.1,speed * (1 - victims_around/coeff_congestion)]);	
		}
		if(myPath!=nil){
			do follow path: myPath speed: actual_speed;
		} else{
			do goto(target);
		}
		if(cell(location).is_exit){
			do die;
		}
	}
	
	action update_path(point new_target) {
		if (target = nil) or (target != new_target) or (myPath = nil) or (myPath.target != new_target){
			target <- new_target;
			myPath <- free_cells path_between (self.location, target);
		}
	}
	
	action update_perceive_area {
		if (vision_amplitude < 180.0) {
			geometry vision_cone <-cone(heading-vision_amplitude,heading+vision_amplitude);
			perceived_area <- vision_cone intersection circle(perceive_distance); 
		} else {
			perceived_area <- circle(perceive_distance);
		}
		if (use_masked_by) {
			perceived_area <- perceived_area masked_by (wall,20);
		}
	}
	
	init{
		location <- one_of(free_cells).location;
		do add_desire(wandering);
		do add_desire(noFire,1.0);
		
		do add_belief(reliableSign);
		do update_perceive_area;
		
		//knowledge about the exact location of the exit door
		int know_exitDoor_lt <- max([1,round(gauss(mean_know_exitDoor,std_know_exitDoor))]);
		do add_belief(predicate:exitDoor, lifetime: know_exitDoor_lt);
		
		do add_belief(predicate:directionExitDoor, lifetime:2*know_exitDoor_lt);
		
		if(!hasFriend and flip(propFriend)){
			friend <- one_of(victim where ((!each.hasFriend)and(each distance_to self <50)));
			if(!(friend=nil)){
				hasFriend<-true;
				do add_social_link(new_social_link(friend,gauss(0.5,0.12),gauss(0.0,0.33),gauss(0.5,0.12),gauss(0.5,0.12)));
				solidarity <- get_solidarity(get_social_link(new_social_link(friend)));
				ask friend{
					friend <- myself;
					hasFriend<-true;
					do add_social_link(new_social_link(friend,gauss(0.5,0.12),gauss(0.0,0.33),gauss(0.5,0.12),gauss(0.5,0.12)));
					solidarity <- get_solidarity(get_social_link(new_social_link(friend)));
				}
			}
		}
		
	}
	
	
	reflex loosingLife{
		if(cell(location).smoke_level>50.0){
			life <- life - 2;
			if(not injured){
				injured <- true;
				nbInjured <- nbInjured + 1;
			}
		}
		if(life<=0){
			nbDead <- nbDead+1;
			do die;
		}
		if(cell(location).is_exit){
			do die;
		}
	}
	
	perceive target: self parallel: true{
		if(smokeQuantity>50){
			perceive_distance <- distance_in_smoke;
			if(not has_belief(exitDoor)){
				needHelp <- true;
			}
		} else {
			perceive_distance <- normal_distance;
			needHelp <- false;
		}
		if(smokeQuantity > max_smoke){
			thresholdLaw <- 0.0;
		}
		smokeQuantity<-0.0;
		do update_perceive_area;
		
	}
	
	perceive target:cell  in:perceived_area parallel: true{
		if(smoke_level>50){
			focus id:"smoke";
			if(myself.perceive_distance=normal_distance){
				if(myself.smokeQuantity<100){myself.smokeQuantity <- myself.smokeQuantity+5;}
			}
			if(myself.perceive_distance=distance_in_smoke){
				if(myself.smokeQuantity<100){myself.smokeQuantity <- myself.smokeQuantity+10;}
			}
		}
	}
		
	perceive target:victim in:perceived_area{
		if(has_belief(fireSaw) and not myself.has_belief(fireSaw)){
			focus id:"fire" strength: uncertaintyConversion is_uncertain:true;
//			ask myself{
//				do add_uncertainty(predicate:fireSaw,strength: uncertaintyConversion);
//			}
		}
	}		
	
	perceive target:victim in:perceived_area parallel:false{
		emotional_contagion emotion_detected:fearFire threshold:contagionThreshold;
		socialize trust:gauss(0.0,0.33);
		myself.perceivedOther<-self;
		enforcement norm: "followOthers" sanction: "trustSanction" reward: "trustReward";
	}
	
	sanction trustSanction{
		do change_trust(perceivedOther,-0.1);
	}
	
	sanction trustReward{
		do change_trust(perceivedOther,0.1);
	}
	
	perceive target:friend in:circle(5.0){
		if(needHelp and self!=myself){
			focus id:"helpFriend" lifetime:30;	
		}
	}
	
	perceive target:exit in:circle(3.5) parallel: true{
		focus id:"exitLocation" lifetime: 20;
	}
	
	perceive target:fire in:circle(5.0) parallel: true{
		focus id:"fire";
	}
	
	perceive target:bathroom in: perceived_area when: has_obligation(fleeing) parallel: true{
		focus id:"wrongExit";
	}
	
	law belief: smokeSaw new_obligation:fleeing when: not has_belief(wrongExit) threshold:thresholdLaw parallel: true;
	
	rule belief: wrongExit remove_obligation: fleeing remove_intention:fleeing;
	rule belief: fireSaw new_desire:fleeing remove_intention:wandering remove_desire:wandering ;
	rule belief: smokeSaw new_uncertainty:fireSaw strength: smokeQuantity/100 when: not has_belief(fireSaw);
	rule emotion:fearFire new_desire:fleeing remove_intention:wandering remove_desire:wandering threshold:riskAversion;
	rule belief: helpingFriend new_desire:helpingFriend remove_intention: fleeing lifetime: 30 strength:1.0 when:(not needHelp) and (agreeableness*solidarity>thresholdFriend);
	
	plan wander intention:wandering{
		speed<-0.5;
		color <- #white;
		if(target=nil){
			target <- one_of(free_cells where (each distance_to self <5.0)).location;
			myPath <- free_cells path_between(location, target);			
		}
		do walk;
		
		if(self distance_to target < (speed * step)){
			target<-nil;
		}
	}

	action walk_to_exit {
		if(target!=first(exit).location){
			do update_path(first(exit).location);
		}
		do walk;
	}
	
	norm followSigns obligation: fleeing when: not has_belief(exitDoor)  finished_when: has_belief(exitDoor) {
		color <- #blue;
		speed<-1.0;
		do update_path(myBathroom.location);
		do walk;
	}
	
	norm followOthers intention: fleeing when:not has_belief(fireSaw) and (smokeQuantity > min_smoke) and (smokeQuantity < max_smoke) and not has_belief(directionExitDoor) finished_when: has_belief(exitDoor) or (smokeQuantity > max_smoke) or (smokeQuantity < min_smoke) threshold:thresholdNorm{
		color <- #brown;
		victim follower <- nil;
		float maxTrust <- 0.0;
		float tempTrust <- -1.0;
		loop tempFollower over: victim where (each distance_to self <perceive_distance){
			tempTrust <- -1.0;
			if(has_social_link(new_social_link(tempFollower))){
				tempTrust <- get_trust(get_social_link(new_social_link(tempFollower)));
			}
			if (tempTrust > maxTrust){
				maxTrust <- tempTrust;
				follower <- tempFollower;
			}
		}
		if(follower != nil){
			target <- follower.target;
			do walk;
		} else {
			cell next_cell <- shuffle(free_cells where (each distance_to self <10.0)) with_min_of (each.smoke_level);
			do update_path(next_cell.location);		
			
			do walk;
			
			if(location = target){
				target<-nil;
			}
		}
	}
	
	plan avoidSmoke intention:fleeing priority: smokeQuantity /5.0 when: not has_belief(exitDoor) and (smokeQuantity > min_smoke) and (smokeQuantity < max_smoke) finished_when: has_belief(exitDoor) or (smokeQuantity > max_smoke) or (smokeQuantity < min_smoke){
		color <- #red;
		speed<-1.5;
		cell next_cell <- shuffle(free_cells where (each distance_to self <10.0)) with_min_of (each.smoke_level);
		do update_path(next_cell.location);		
		
		do walk;
		
		if(location = target){
			target<-nil;
		}
	}
	
	
	plan toTheExitDirection intention:fleeing when: has_belief(directionExitDoor) and (smokeQuantity < max_smoke) finished_when: (smokeQuantity > max_smoke) or has_belief(exitDoor) priority: 10{
		color <- #yellow;
		speed<-1.5;
		target <- (free_cells where (each distance_to self <10.0) closest_to first(exit).location).location;
		myPath<-free_cells path_between (location, target);
		do walk;
		
		if(self distance_to target < (speed * step) ){
			target<-nil;
		}
	}
	
	plan toTheExit intention:fleeing when: has_belief(exitDoor) and smokeQuantity<min_smoke  finished_when: (smokeQuantity > min_smoke) or (not has_belief(exitDoor)) priority: 20{
		speed<-1.5;
		color <- #green;
		do walk_to_exit;
	}
	
	plan throughTheSmoke intention:fleeing when: has_belief(exitDoor) and smokeQuantity>=max_smoke  finished_when: not has_belief(exitDoor) or smokeQuantity<max_smoke priority: 20{ 
		color <- #purple;
		do walk_to_exit;
	}
	
	plan helpAFriend intention:helpingFriend{
		if(not dead(friend)){
			if (self distance_to friend <= 1.0 #m ) {
						 ask friend{
						 	do add_belief(predicate:exitDoor, lifetime: 10);
						 }
						 do remove_belief(get_predicate(get_current_intention()));
						 do remove_intention(get_predicate(get_current_intention()), true);
				} else {
					speed <- 1.5;
					do update_path(friend.location);
					do walk;
				}
			}
	}
	
	plan wanderInSmoke intention: fleeing when:(smokeQuantity > max_smoke and has_belief(wrongExit)) finished_when: (smokeQuantity < max_smoke) or has_belief(exitDoor) priority:1{
		color <- #darkred;
		target <- (one_of(free_cells where (each distance_to self <10.0))).location;
		myPath<-free_cells path_between (location, target);
		do walk;
		
		if(location = target){
			target<-nil;
		}
	}
	
	aspect default {
		draw triangle(0.25) rotate: 90 + heading color: color border: #black;
	}
	aspect perception {
		if (perceived_area != nil) {
			draw perceived_area color: #green empty:true border:#red ;
		}
	}
}

experiment KissNightclub type:gui{
	output{
		display map /*type: opengl*/{
			overlay position: { 0, 0 } size: { 230 #px, 260 #px } background: # black transparency: 0.5 border: #black rounded: true
            {
            	float y <- 30#px;
                draw triangle(10#px) at: { 20#px, 30#px } color:#white border:#black;
                draw "dance" at: { 40#px, 30#px + 4#px } color: # white font: font("SansSerif", 18, #bold);
                draw triangle(10#px) at: { 20#px, 55#px } color:#green border:#black;
                draw "to the exit" at: { 40#px, 55#px + 4#px } color: # white font: font("SansSerif", 18, #bold);
                draw triangle(10#px) at: { 20#px, 80#px } color:#yellow border:#black;
                draw "to the exit direction" at: { 40#px, 80#px + 4#px } color: # white font: font("SansSerif", 18, #bold);
                draw triangle(10#px) at: { 20#px, 105#px } color:#purple border:#black;
                draw "through the smoke" at: { 40#px, 105#px + 4#px } color: # white font: font("SansSerif", 18, #bold);
                draw triangle(10#px) at: { 20#px, 130#px } color:#red border:#black;
                draw "avoid smoke" at: { 40#px, 130#px + 4#px } color: # white font: font("SansSerif", 18, #bold);
                draw triangle(10#px) at: { 20#px, 155#px } color:#orange border:#black;
                draw "help a friend" at: { 40#px, 155#px + 4#px } color: # white font: font("SansSerif", 18, #bold);
                draw triangle(10#px) at: { 20#px, 180#px } color:#blue border:#black;
                draw "follow sign" at: { 40#px, 180#px + 4#px } color: # white font: font("SansSerif", 18, #bold);
                draw triangle(10#px) at: { 20#px, 205#px } color:#brown border:#black;
                draw "follow others" at: { 40#px, 205#px + 4#px } color: # white font: font("SansSerif", 18, #bold);
                draw triangle(10#px) at: { 20#px, 230#px } color:#darkred border:#black;
                draw "wander in smoke" at: { 40#px, 230#px + 4#px } color: # white font: font("SansSerif", 18, #bold);

            }
			
			grid cell lines: #white;
			species wall refresh: false;
			species exit refresh: false;
			species victim;
			species fire;
		}
		
		display charts {
			chart "causalities" {
				data "nb people" value: length(victim) color: #blue;
				data "nb dead" value: nbDead color: #red;
				data "nb outside" value: nb_victims - length(victim) - nbDead color: #green;
				data "nb injured" value: nbInjured-nbDead color: #yellow;
			}
		}
	}
}

experiment kissBatch type: batch repeat:16{
	reflex infos{
		write "nbDead " + nbDead;
	}
}

